from typing import (
    TYPE_CHECKING,
    Awaitable,
    overload,
    Optional,
    Iterator,
    Callable,
    TypeVar,
    Union,
    Tuple,
    Type,
    List,
    Dict,
    cast,
)
from concurrent.futures import CancelledError
import threading
import asyncio

from typing_extensions import Literal
from jinja2.nodes import Template

from lona.view_runtime import VIEW_RUNTIME_STATE, ViewRuntime
from lona.exceptions import ServerStop, UserAbort
from lona.html.abstract_node import AbstractNode
from lona.events.input_event import InputEvent
from lona.shell.shell import embed_shell
from lona.connection import Connection
from lona.errors import ClientError
from lona.request import Request

# avoid import cycles
if TYPE_CHECKING:
    from lona.server import LonaServer


_T = TypeVar('_T')
_V = TypeVar('_V', bound='LonaView')
_HTML = Union[None, AbstractNode, str]


class LonaView:
    _objects: Dict[str, List['LonaView']] = {}

    def __init__(
            self,
            server: 'LonaServer',
            view_runtime: ViewRuntime,
            request: Request,
    ) -> None:
        self._server: 'LonaServer' = server
        self._view_runtime: ViewRuntime = view_runtime
        self._request: Request = request

    # objects #################################################################
    @overload
    @classmethod
    def _get_objects_key(
            cls,
            view_class: Type['LonaView'],
            is_class: Literal[True],
    ) -> str:
        ...

    @overload
    @classmethod
    def _get_objects_key(
            cls,
            view_class: 'LonaView',
            is_class: Literal[False],
    ) -> str:
        ...

    @classmethod
    def _get_objects_key(cls, view_class, is_class):
        if not is_class:
            view_class = view_class.__class__

        return f'{view_class.__module__}.{view_class.__name__}'

    @classmethod
    def _add_view_to_objects(cls, view_object: 'LonaView') -> None:
        objects_key = cls._get_objects_key(view_object, is_class=False)

        if objects_key not in cls._objects:
            cls._objects[objects_key] = []

        cls._objects[objects_key].append(view_object)

    @classmethod
    def _remove_view_from_objects(cls, view_object: 'LonaView') -> None:
        objects_key = cls._get_objects_key(view_object, is_class=False)

        if(objects_key in cls._objects and
           view_object in cls._objects[objects_key]):

            cls._objects[objects_key].remove(view_object)

    @classmethod
    def iter_objects(cls: Type[_V]) -> Iterator[_V]:
        objects_key = cls._get_objects_key(cls, is_class=True)
        view_objects = cls._objects.get(objects_key, []).copy()
        yield from view_objects  # type: ignore

    # properties ##############################################################
    @property
    def server(self) -> 'LonaServer':
        return self._server

    @property
    def request(self) -> Request:
        return self._request

    # checks ##################################################################
    def _assert_not_main_thread(self) -> None:
        if threading.current_thread() is threading.main_thread():
            raise RuntimeError(
                'this function is not supposed to run in the main thread')

    def _assert_view_is_interactive(self) -> None:
        if not self._request.route.interactive:
            raise RuntimeError(
                'operation is not supported in non-interactive requests')

    def _assert_view_is_running(self) -> None:
        if self._view_runtime.stop_reason:
            raise self._view_runtime.stop_reason

    # asyncio #################################################################
    def _await_sync(self, awaitable: Awaitable[_T]) -> _T:
        async def await_awaitable() -> _T:
            finished, pending = await asyncio.wait(
                [
                    cast(Awaitable, self._view_runtime.stopped),
                    awaitable,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            for finished_future in finished:
                if finished_future is not self._view_runtime.stopped:
                    return finished_future.result()

            if self._view_runtime.stopped in finished:
                for pending_future in pending:
                    if not pending_future.done():
                        pending_future.cancel()

            raise self._view_runtime.stop_reason

        return asyncio.run_coroutine_threadsafe(
            await_awaitable(),
            loop=self._server.loop,
        ).result()

    def await_sync(self, awaitable: Awaitable[_T]) -> _T:
        self._view_runtime.state = VIEW_RUNTIME_STATE.WAITING_FOR_IOLOOP

        try:
            return self._await_sync(awaitable)

        finally:
            self._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    # html ####################################################################
    def show(
            self,
            html: _HTML = None,
            template: Union[None, str, Template] = None,
            template_string: Union[None, str, Template] = None,
            title: Optional[str] = None,
            template_context: Optional[dict] = None,
    ) -> None:
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        # templating
        if template or template_string:
            template_context = template_context or {}

            if 'template_context' in template_context:
                template_context = template_context['template_context']

            # string based templates
            if template_string:
                html = self._server.templating_engine.render_string(
                    template_string=template_string,
                    template_context=template_context,
                )

            # file based templates
            else:
                html = self._server.templating_engine.render_template(
                    template_name=template,
                    template_context=template_context,
                )

        with self._view_runtime.document.lock:
            html = html or self._view_runtime.document.html
            data = self._view_runtime.document.apply(html)

            if data:
                self._view_runtime.send_data(data=data, title=title)

    def set_title(self, title: str) -> None:
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        with self._view_runtime.document.lock:
            self._view_runtime.send_data(title=title)

    # messaging ###############################################################
    def send_str(
            self,
            string: str,
            broadcast: bool = False,
            filter_connections: Callable[[Connection], bool] = lambda c: True,
            wait: bool = True,
    ) -> None:
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        # broadcast
        if broadcast:
            for connection in self.server.websocket_connections.copy():
                if not filter_connections(connection):
                    continue

                connection.send_str(string, wait=wait)

        # view local
        else:
            runtime = self._view_runtime

            with runtime.document.lock:
                for connection in list(runtime.connections.keys()):
                    if not connection.interactive:
                        continue

                    connection.send_str(string, wait=wait)

    # input events ############################################################
    def _await_specific_input_event(
            self,
            *nodes: Union[AbstractNode, List[AbstractNode]],
            event_type: str = '',
            html: _HTML = None,
    ) -> InputEvent:
        self._view_runtime.state = VIEW_RUNTIME_STATE.WAITING_FOR_INPUT

        try:
            self._assert_not_main_thread()
            self._assert_view_is_interactive()
            self._assert_view_is_running()

            if len(nodes) == 1 and isinstance(nodes[0], list):
                nodes = tuple(nodes[0])

            nodes = cast(Tuple[AbstractNode], nodes)

            if html is not None:
                self.show(html=html)

            return self._view_runtime.await_input_event(
                nodes=nodes,
                event_type=event_type,
            )

        finally:
            self._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    @overload
    def await_input_event(self, *nodes: AbstractNode, html: _HTML = None) -> InputEvent:  # NOQA: LN001
        ...

    @overload
    def await_input_event(self, __nodes: List[AbstractNode], html: _HTML = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_input_event(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='event',
            html=html,
        )

    @overload
    def await_click(self, *nodes: AbstractNode, html: _HTML = None) -> InputEvent:  # NOQA: LN001
        ...

    @overload
    def await_click(self, __nodes: List[AbstractNode], html: _HTML = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_click(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='click',
            html=html,
        )

    @overload
    def await_change(self, *nodes: AbstractNode, html: _HTML = None) -> InputEvent:  # NOQA: LN001
        ...

    @overload
    def await_change(self, __nodes: List[AbstractNode], html: _HTML = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_change(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='change',
            html=html,
        )

    # runtime #################################################################
    @overload
    def sleep(self, delay: float) -> None:
        ...

    @overload
    def sleep(self, delay: float, result: _T) -> _T:
        ...

    def sleep(self, delay: float, result: Optional[_T] = None) -> Optional[_T]:
        self._view_runtime.state = VIEW_RUNTIME_STATE.SLEEPING

        try:
            return self._await_sync(asyncio.sleep(delay, result))

        finally:
            self._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    def daemonize(self) -> None:
        self._assert_view_is_interactive()

        self._view_runtime.is_daemon = True

    def ping(self) -> Literal['pong']:
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        return 'pong'

    # helper ##################################################################
    def embed_shell(self, _locals: Optional[dict] = None) -> None:
        if _locals is None:
            _locals = {}
        _locals['self'] = self

        embed_shell(server=self.server, locals=_locals)

    # hooks ###################################################################
    def handle_request(self, request: Request) -> Union[None, str, AbstractNode, dict]:  # NOQA: LN001
        return ''

    def handle_input_event_root(self, input_event: InputEvent) -> Optional[InputEvent]:  # NOQA: LN001
        return input_event

    def handle_input_event(self, input_event: InputEvent) -> Optional[InputEvent]:  # NOQA: LN001
        return input_event

    def on_shutdown(
            self,
            reason: Union[
                None,
                UserAbort,
                ServerStop,
                CancelledError,
                ClientError,
                Exception,
            ],
    ) -> None:
        pass



from typing import overload, Optional, Callable, Union, Type, List, Dict, Any
from tempfile import TemporaryDirectory
from asyncio import AbstractEventLoop
from os import PathLike
import logging
import os

from typing_extensions import Literal
from aiohttp.web import Application

from lona.command_line.run_server import run_server
from lona import default_settings, LonaView
from lona.worker_pool import WorkerPool
from lona.logging import setup_logging
from lona.settings import Settings
from lona.server import LonaServer
from lona.routing import Route

logger = logging.getLogger('lona.app')


class ServerArgs:
    def __init__(self, **args):
        self.add(**args)

    def add(self, **args):
        for key, value in args.items():
            setattr(self, key, value)


class LonaApp:
    def __init__(self, script_path: PathLike) -> None:
        self.script_path: PathLike = script_path
        self.project_root: str = os.path.dirname(self.script_path)

        self.aiohttp_app: Optional[Application] = None
        self.server: Optional[LonaServer] = None

        # setup tempdir
        self.temp_dir: TemporaryDirectory = TemporaryDirectory()

        self.template_dir: str = os.path.join(
            self.temp_dir.name,
            'templates',
        )

        self.static_dir: str = os.path.join(
            self.temp_dir.name,
            'static',
        )

        os.makedirs(self.template_dir)
        os.makedirs(self.static_dir)

        # setup routes
        self.routes: List[Route] = []

        # setup settings
        self.settings: Settings = Settings()
        self.settings.add(default_settings.__file__)

    # helper ##################################################################
    def _get_settings_as_dict(self) -> Dict[str, Any]:
        settings = {}

        for name in self.settings:
            settings[name] = self.settings.get(name)

        return settings

    def resolve_path(self, path: str) -> str:
        if path.startswith('/'):
            return path

        return os.path.normpath(
            os.path.join(
                self.project_root,
                path,
            ),
        )

    # decorator ###############################################################
    def route(
            self,
            # 1 = lona.MATCH_ALL  https://github.com/python/mypy/issues/10026
            raw_pattern: Union[str, Literal[1]],
            name: str = '',
            interactive: bool = True,
            http_pass_through: bool = False,
            frontend_view: Union[None, str, LonaView] = None,
    ) -> Callable[[Type[LonaView]], None]:
        def decorator(view_class: Type[LonaView]) -> None:
            self.routes.append(
                Route(
                    raw_pattern=raw_pattern,
                    view=view_class,
                    name=name,
                    interactive=interactive,
                    http_pass_through=http_pass_through,
                    frontend_view=frontend_view,
                ),
            )

        return decorator

    @overload
    def middleware(self) -> Callable[[Type], None]:
        ...

    @overload
    def middleware(self, middleware_class: Type) -> None:
        ...

    def middleware(self, middleware_class: Optional[Type] = None) -> Optional[Callable[[Type], None]]:  # NOQA: LN001
        def decorator(middleware_class: Type) -> None:
            self.settings.MIDDLEWARES.append(middleware_class)

        if callable(middleware_class):
            decorator(middleware_class)
            return None

        else:
            return decorator

    @overload
    def frontend_view(self) -> Callable[[Type[LonaView]], None]:
        ...

    @overload
    def frontend_view(self, view_class: Type[LonaView]) -> None:
        ...

    def frontend_view(self, view_class: Optional[Type[LonaView]] = None) -> Optional[Callable[[Type[LonaView]], None]]:  # NOQA: LN001
        def decorator(view_class: Type[LonaView]) -> None:
            self.settings.FRONTEND_VIEW = view_class

        if callable(view_class):
            decorator(view_class)
            return None

        else:
            return decorator

    # files ###################################################################
    def _add_file(
            self,
            temp_dir: str,
            name: str,
            string: str = '',
            path: str = '',
    ) -> None:
        if name.startswith('/'):
            name = name[1:]

        dirname = os.path.join(
            temp_dir,
            os.path.dirname(name),
        )

        try:
            os.makedirs(dirname)

        except FileExistsError:
            pass

        full_name = os.path.join(
            dirname,
            os.path.basename(name),
        )

        if path:
            _path = self.resolve_path(path)

            if not os.path.exists(_path):
                raise FileNotFoundError(path)

            os.symlink(os.path.abspath(_path), full_name)

        elif string:
            with open(full_name, 'w+') as f:
                f.write(string)

        else:
            raise ValueError('either string or path have to be set')

    def add_template(
            self,
            name: str,
            string: str = '',
            path: str = '',
    ) -> None:
        return self._add_file(
            temp_dir=self.template_dir,
            name=name,
            string=string,
            path=path,
        )

    def add_static_file(
            self,
            name: str,
            string: str = '',
            path: str = '',
    ) -> None:
        return self._add_file(
            temp_dir=self.static_dir,
            name=name,
            string=string,
            path=path,
        )

    # server ##################################################################
    def setup_server(
            self,
            loop: Union[AbstractEventLoop, None] = None,
    ) -> None:

        # finish settings
        self.settings.CORE_TEMPLATE_DIRS.insert(0, self.template_dir)
        self.settings.STATIC_DIRS.insert(0, self.static_dir)

        # setup server
        self.aiohttp_app = Application(loop=loop)

        self.server = LonaServer(
            app=self.aiohttp_app,
            project_root=self.project_root,
            settings_post_overrides=self._get_settings_as_dict(),
            routes=self.routes,
        )

        # setup worker pool
        worker_pool = WorkerPool(
            settings=self.server.settings,
        )

        self.server.set_loop(loop)
        self.server.set_worker_pool(worker_pool)

    def run(
            self,
            loop: Union[AbstractEventLoop, None] = None,
            **args: Any,
    ) -> None:

        self.setup_server(loop=loop)

        # setup arguments
        server_args = ServerArgs(
            host='localhost',
            port=8080,
            shell_server_url='',
            shutdown_timeout=0,
            log_level='info',
            loggers=[],
            debug_mode='',
            shell=False,
        )

        server_args.add(**args)

        setup_logging(server_args)

        # start server
        run_server(
            args=server_args,
            app=self.aiohttp_app,
            server=self.server,
        )



from aiohttp.web import Application
import pytest

from lona.worker_pool import WorkerPool
from lona.server import LonaServer
from lona import LonaApp


class LonaContext:
    def __init__(self, client, app, server):
        self.client = client
        self.app = app
        self.server = server

    def make_url(self, path=''):
        return str(self.client.make_url(path))


@pytest.fixture()
def lona_app_context(request, aiohttp_client):
    async def setup_lona_app_context(setup_app, project_root=''):
        # setup lona app
        lona_app = LonaApp(project_root or request.fspath)
        setup_app(lona_app)

        # setup client
        def setup_aiohttp_app(loop):
            lona_app.setup_server(loop=loop)

            return lona_app.aiohttp_app

        client = await aiohttp_client(setup_aiohttp_app)

        return LonaContext(
            client=client,
            app=lona_app,
            server=lona_app.server,
        )

    return setup_lona_app_context


@pytest.fixture()
def lona_project_context(request, aiohttp_client):
    async def setup_lona_project_context(
        project_root='',
        settings=None,
        settings_pre_overrides=None,
        settings_post_overrides=None,
    ):

        # setup aiohttp app
        server = None

        def setup_aiohttp_app(loop):
            nonlocal server

            aiohttp_app = Application(loop=loop)

            # setup lona server
            server = LonaServer(
                app=aiohttp_app,
                project_root=project_root or request.fspath,
                settings_paths=settings or [],
                settings_pre_overrides=settings_pre_overrides or {},
                settings_post_overrides=settings_post_overrides or {},
            )

            server.set_loop(loop)
            server.set_worker_pool(WorkerPool(settings=server.settings))

            return aiohttp_app

        client = await aiohttp_client(setup_aiohttp_app)

        return LonaContext(
            client=client,
            app=None,
            server=server,
        )

    return setup_lona_project_context
